/* $Id: gdisplay.c,v 1.1 2007/07/15 22:05:23 mgdepalma Exp $
 *
 * Copyright (C) 2007 Softcraft Impresa, LLC <info@softcraft.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; You may only use version 2 of the License,
 * you have no option to use any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "gdisplay.h"
#include "language.h"

#ifndef get_current_dir_name
extern char *get_current_dir_name(void);
#endif

static gdisplay_t global;	/* (protected) encapsulated program data */

const char *Program;		/* (public) published program name */
const char *Release;		/* (public) published program version */

/*
 * (private) about
 */
static gboolean
about(GtkWidget *instance, gpointer data)
{
  static char *Information =
  "is a gtk-based simple image viewing utility and directory browser.\n"
  "The image selected can be used to set the screen background and save\n"
  "the setting in $HOME/.config/desktop used by Generations Linux.\n"
  "\n"
  "The program is developed by Mauro DePalma and distributed\n"
  "under the terms and condition of the GNU Public License.\n";

  notice(global->window, ICON_SNAPSHOT,
                 strargs("\n%s %s %s", Program, Release, Information));
  return FALSE;
} /* </about> */

/*
 * clean application exit
 */
static gboolean
finis(GtkWidget *instance, gpointer data)
{
  gtk_main_quit ();
  return FALSE;
} /* </finis> */

/*
 * refresh view of the canvas area
 */
static void
refresh (GtkWidget *canvas, gpointer data)
{
  GError    *error = NULL;
  GdkPixbuf *image = NULL;
  gchar     *name;

  if ( (name = filechooser_get_selected_name (global->chooser)) )
    image = gdk_pixbuf_new_from_file (name, &error);

  if (image) {
    /* redraw_pixbuf() always clears the drawing area */
    redraw_pixbuf (canvas, image);
    gdk_pixbuf_unref (image);

    if (global->background)
      gtk_widget_set_sensitive(GTK_WIDGET(global->apply), TRUE);
  }
  else {
    if (name) {
      redraw_pixbuf (canvas, image = xpm_pixbuf(ICON_BROKEN, NULL));
      gdk_pixbuf_unref (image);
    }
    gtk_widget_set_sensitive(GTK_WIDGET(global->apply), FALSE);
    /* printf("%s: %s\n", Program, error->message); */
  }
  gtk_widget_set_sensitive(GTK_WIDGET(global->save), FALSE);
} /* </refresh> */

/*
 * (private) agent callback function
 */
static gboolean
agent(gpointer data)
{
  struct stat info;

  if (lstat((gchar *)data, &info) == 0 && S_ISREG(info.st_mode))
    refresh(global->canvas, NULL);

  return TRUE;
} /* </agent> */

/*
 * set the value for using image as background
 */
static void
background(GtkToggleButton *button)
{
  global->background = button->active;
  gtk_widget_set_sensitive(GTK_WIDGET(global->apply), global->background);

  if (global->background == FALSE)
    gtk_widget_set_sensitive(GTK_WIDGET(global->save), FALSE);
} /* </background> */

/*
 * (private) changemode callback function
 */
static gboolean
changemode(GtkWidget *button, gpointer data)
{
  GtkWidget *frame  = global->window;
  GdkWindow *window = frame->window;

  gint width, height;	/* width and height of resulting global->window */

  /* Toggle with and without global->browser */
  if (global->filer) {
    GError    *error = NULL;
    GdkPixbuf *image = NULL;
    gchar     *name;

    if ( (name = filechooser_get_selected_name (global->chooser)) )
      image = gdk_pixbuf_new_from_file (name, &error);

    if (image) {
      gdk_pixbuf_unref (image);	/* image out of scope */

      width  = 500;		/* proportional width and height */
      height = 360;

      gtk_widget_hide (global->browser);
      gtk_widget_set_usize (frame, width, height);
      gdk_window_resize (window, width, height);

      gtk_button_set_image (GTK_BUTTON(button), xpm_image(ICON_FOLDER));
      global->filer = FALSE;
    }
  }
  else {
    width  = global->width;	/* original width and height */
    height = global->height;

    gtk_widget_show (global->browser);
    gtk_widget_set_usize (frame, width, height);
    gdk_window_resize (window, width, height);

    gtk_button_set_image (GTK_BUTTON(button), xpm_image(ICON_CHOOSER));
    global->filer = TRUE;
  }
  return TRUE;
} /* </changemode> */

/*
 * (private) prevfile
 * (private) nextfile
 */
static gboolean
prevfile(GtkWidget *button, FileChooser *chooser)
{
  int index = filechooser_get_cursor (chooser);

  if (index > 0) {
    filechooser_set_cursor (chooser, --index);
    refresh(global->canvas, NULL);
  }
  return TRUE;
} /* </prevfile> */

static gboolean
nextfile(GtkWidget *button, FileChooser *chooser)
{
  int index = filechooser_get_cursor (chooser);
  int count = filechooser_get_count (chooser);

  if (index < count) {
    filechooser_set_cursor (chooser, ++index);
    refresh(global->canvas, NULL);
  }
  return TRUE;
} /* </nextfile> */

/*
 * (private) set background callback
 */
static gboolean
setbg(GtkWidget *widget, gpointer data)
{
  gchar *pathname = filechooser_get_selected_name (global->chooser);

  if (pathname) {
    GError    *error  = NULL;
    GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file (pathname, &error);

    if (pixbuf) {
      set_background_pixbuf (gdk_get_default_root_window(), pixbuf);
      gdk_pixbuf_unref (pixbuf);

      /* Reset APPLY and SAVE button state */
      gtk_widget_set_sensitive(GTK_WIDGET(global->apply), FALSE);
      gtk_widget_set_sensitive(GTK_WIDGET(global->save), TRUE);
    }
    else {
      /* show error dialog */
    }
  }
  return TRUE;
} /* </setbg> */

/*
 * (private) configsave
 */
static gboolean
configsave(GtkWidget *widget, gpointer data)
{
  FILE *stream;
  char **buffer = NULL;
  char line[MAXPATH];
  int count = 0;
  int idx;

  /* Read global->resource in memory buffer. */
  if ((stream = fopen(global->resource, "r"))) {
    while (fgets(line, MAXPATH, stream)) ++count;

    /* Allocate memory space for global->resource */
    buffer = malloc(count * sizeof (char *));
    fseek(stream, 0L, SEEK_SET);

    for (idx=0; idx < count; idx++) {
      if (fgets(line, MAXPATH, stream) <= 0) break;
      buffer[idx] = strdup(line);
    }
    fclose(stream);
  }

  /* [Re]open global->resource for writing. */
  if ((stream = fopen(global->resource, "w"))) {
    gboolean insert = TRUE;
    gchar *pathname = filechooser_get_selected_name (global->chooser);

    for (idx=0; idx < count; idx++) {
      if (strncmp(buffer[idx], "BACKGROUND=", 11) == 0) {
        fprintf(stream, "BACKGROUND=\"%s\"\n", pathname);
        insert = FALSE;
      }
      else {
        fputs(buffer[idx], stream);
      }
    }
    /* "BACKGROUND=" was not specified in global->resource */
    if (insert) fprintf(stream, "BACKGROUND=\"%s\"\n", pathname);

    fclose(stream);
  }

  /* Free the memory space allocated for global->resource */
  if (buffer) {
    for (idx=0; idx < count; idx++) free(buffer[idx]);
    free(buffer);
  }

  /* Disable the global->save button. */
  gtk_widget_set_sensitive(GTK_WIDGET(global->save), FALSE);

  return TRUE;
} /* </configsave> */

/*
 * interface - construct user interface
 */
static GtkWidget*
interface (GtkWidget *window)
{
  FileChooser *chooser = global->chooser;

  GtkWidget *layout = gtk_vbox_new (FALSE, 0);
  GtkWidget *area, *box, *button, *field, *inset;
  GtkWidget *canvas, *frame, *scroll, *split;


  /* Split view: file chooser on the left and preview on the right. */
  split = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX (layout), split, TRUE, TRUE, 0);
  gtk_widget_show(split);

  /* Assemble scrollable file chooser. */
  area = global->browser = gtk_vbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX (split), area, TRUE, TRUE, 0);
  gtk_widget_show(area);

  /* Assemble box containing navigation and current directory. */
  box = chooser->dirbox;
  gtk_box_pack_start(GTK_BOX (area), box, FALSE, FALSE, 2);
  gtk_widget_show(box);

  /* Assemble file selector. */
  scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start(GTK_BOX (area), scroll, TRUE, TRUE, 5);
  gtk_widget_show(scroll);

  area = chooser->viewer;
  gtk_container_add (GTK_CONTAINER(scroll), area);
  gtk_widget_show(area);

  /* Neutral zone between global->browser and image viewer */
  inset = gtk_label_new ("");
  gtk_box_pack_start(GTK_BOX (split), inset, FALSE, FALSE, 4);
  gtk_widget_show(inset);

  /* Assemble display canvas area. */
  area = global->viewer = gtk_vbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX (split), area, TRUE, TRUE, 0);
  gtk_widget_show(area);

  /* Previous and next selection navigation. */
  box = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX (area), box, FALSE, TRUE, 0);
  gtk_widget_show(box);

  button = global->backward = xpm_button(ICON_BACK, NULL);
  gtk_box_pack_start(GTK_BOX (box), button, FALSE, TRUE, 0);
  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(prevfile), chooser);
  gtk_widget_show(button);

  button = global->forward = xpm_button(ICON_FORWARD, NULL);
  gtk_box_pack_start(GTK_BOX (box), button, FALSE, TRUE, 0);
  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(nextfile), chooser);
  gtk_widget_show(button);

  /* Program information (about) button. */
  button = xpm_button(ICON_HELP, _L(LABEL_INFO));
  gtk_box_pack_end (GTK_BOX(box), button, FALSE, TRUE, 2);
  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
  g_signal_connect (G_OBJECT(button), "clicked", G_CALLBACK(about), NULL);
  gtk_widget_show (button);

  /* Image preview canvas drawing area. */
  canvas = global->canvas = gtk_drawing_area_new ();
  gtk_box_pack_start(GTK_BOX (area), canvas, TRUE, TRUE, 5);

  g_signal_connect (G_OBJECT (canvas), "expose_event",
                           G_CALLBACK (refresh), NULL);
  gtk_widget_show(canvas);

  /* Set as background check box. */
  button = gtk_check_button_new_with_label(_L(LABEL_SET_BACKGROUND));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), FALSE);
  gtk_box_pack_start(GTK_BOX (area), button, FALSE, FALSE, 2);

  g_signal_connect (G_OBJECT (button), "toggled",
                          G_CALLBACK (background), NULL);
  gtk_widget_show(button);

  /* Bottom layout area: display file name and control buttons */
  area = chooser->namebox;
  gtk_box_pack_start(GTK_BOX(layout), area, FALSE, TRUE, 0);
  gtk_widget_show(area);

  /* Neutral zone between global->browser and global->window right border */
  inset = gtk_label_new ("");
  gtk_box_pack_start(GTK_BOX (split), inset, FALSE, FALSE, 4);
  gtk_widget_show(inset);

  /* (global->chooser)->actuator to open and close global->browser */
  button = chooser->actuator;
  gtk_button_set_image (GTK_BUTTON(button), xpm_image(ICON_CHOOSER));
  g_signal_connect(G_OBJECT (button), "clicked", G_CALLBACK(changemode), NULL);

  /* (global->chooser)->name readonly entry to display file name */
  field = chooser->name;
  gtk_entry_set_editable (GTK_ENTRY(field), FALSE);
  gtk_widget_set_usize(GTK_WIDGET(field), global->width/3, 0);

  /* APPLY, SAVE and CLOSE buttons */
  box = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end(GTK_BOX (area), box, FALSE, FALSE, 0);
  gtk_widget_show(box);

  /* APPLY button */
  button = global->apply = xpm_button(ICON_APPLY, _L(LABEL_APPLY));
  g_signal_connect (G_OBJECT(button), "clicked", G_CALLBACK(setbg), NULL);
  gtk_box_pack_start(GTK_BOX (box), button, FALSE, FALSE, 0);

  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
  gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);
  gtk_widget_show(button);

  /* Save resource configuration changes button. */
  button = global->save = xpm_button(ICON_SAVE, _L(LABEL_SAVE));
  g_signal_connect (G_OBJECT(button), "clicked", G_CALLBACK(configsave), NULL);
  gtk_box_pack_start(GTK_BOX (box), button, FALSE, FALSE, 0);

  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
  gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);
  gtk_widget_show(button);

  /* CLOSE button */
  button = xpm_button(ICON_CLOSE, _L(LABEL_CLOSE));
  g_signal_connect (G_OBJECT(button), "clicked", G_CALLBACK(finis), NULL);
  gtk_box_pack_start(GTK_BOX (box), button, FALSE, FALSE, 0);

  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
  gtk_widget_show(button);

  return layout;
} /* </interface> */

/*
 * initialize
 */
static void
initialize (gdisplay_t self, char *imagepath)
{
  static char path[MAXPATH];  /* dirname(1) of the BACKGROUND resource */

  char line[MAXPATH];
  char *resource, *scan;
  char *home = getenv("HOME");
  char *name = NULL;
  int error = 0;
  int mark;

  FileChooser *chooser;
  FILE *stream;


  /* global program data structure */
  global = self;
  global->background = FALSE;
  global->filer = TRUE;		/* controls showing the file chooser */

  global->width  = (6 * gdk_screen_width ())  / 9;
  global->height = (4 * gdk_screen_height ()) / 9;

  /* $HOME/.config/desktop is the user specific resource file */
  sprintf(global->resource, "%s/.config/desktop", home);

  /* /etc/sysconfig/desktop is the system wide resource file */
  if (access(global->resource, R_OK) == 0) {
    resource = global->resource;
  }
  else {
    resource = "/etc/sysconfig/desktop";
    sprintf(line, "%s/.config", home);
    mkdir(line, 0755);
  }

  /* Program called with an image file name. */
  if (imagepath) {
    if (access(imagepath, R_OK) != 0) {
      fprintf(stderr, "%s: no such file or directory\n", imagepath);
      error = ENOENT;
    }
    else {
      if (imagepath[0] == '/')
        strcpy(path, imagepath);
      else
        sprintf(path, "%s/%s", get_current_dir_name(), imagepath);
    }
  }
  else if ((stream = fopen(resource, "r"))) {
    /* Search for "BACKGROUND=<file name>" in the resource file. */

    while (fgets(line, MAXPATH, stream)) {
      if (strncmp(line, "BACKGROUND=", 11) == 0) {
        if (line[11] == '"') {
          line[strlen(line) - 2] = (char)0;
          mark = 12;
        }
        else {
          line[strlen(line) - 1] = (char)0;
          mark = 11;
        }
        strcpy(path, &line[mark]);
	break;
      }
    }
    fclose(stream);
  }

  /* fallback to the user home directory */
  if (error || access(path, R_OK) != 0) {
    strcpy(path, home);
  }
  else if ( (scan = strrchr(path, '/')) ) {
    *scan = (char)0;
    name = ++scan;
  }

  /* Make sure we set (global->chooser)->agent */
  chooser = global->chooser = filechooser_new (path);
  chooser->agent = agent;

  if (name) {
    int index = filechooser_get_index (chooser, name);
    if (index >= 0) filechooser_set_cursor (chooser, index);
    gtk_entry_set_text (GTK_ENTRY(chooser->name), name);
  }
} /* </initialize> */

/*
 * main - main program implementation
 */
int main(int argc, char *argv[])
{
  struct _gdisplay memory;	/* see, gdisplay.h */
  GtkWidget *window;		/* main window     */
  GtkWidget *layout;		/* user interface  */

  Program = basename(argv[0]);
  Release = "1.0";

  /* Initialization of the GTK */
  gtk_set_locale ();
  gtk_init (&argc, &argv);

  /* Initialize program data structure */
  initialize (&memory, (argc > 1) ? argv[1] : NULL);

  /* Create a new top level window */
  window = global->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  /* Set title and border width of the main window */
  gtk_window_set_title (GTK_WINDOW (window), Program);
  gtk_window_set_icon (GTK_WINDOW (window), xpm_pixbuf(ICON_CHOOSER, NULL));
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_widget_set_usize (window, global->width, global->height);
  gtk_widget_show (window);

  /* Construct the user interface */
  layout = interface (window);
  gtk_container_add (GTK_CONTAINER(window), layout);
  gtk_widget_show (layout);

  /* Set event handlers for window close/destroy */
  g_signal_connect (G_OBJECT (window), "destroy",
                      G_CALLBACK (finis), NULL);

  g_signal_connect (G_OBJECT (window), "delete_event",
                      G_CALLBACK (finis), NULL);

  /* See if we were called with a file to display */
  if (argc > 1 && access(argv[1], R_OK) == 0)
    changemode ((global->chooser)->actuator, NULL);

  /* Main event loop */
  gtk_main ();
    
  return Success;
} /* </main> */
