/* $Id: window.c,v 1.1 2007/07/15 22:05:23 mgdepalma Exp $
 *
 * Copyright (C) 2007 Softcraft Impresa, LLC <info@softcraft.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; You may only use version 2 of the License,
 * you have no option to use any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "window.h"

/*
 * scale a GdkPixbuf to given width and height
 */
GdkPixbuf *
pixbuf_scale(GdkPixbuf *image, gint width, gint height)
{
  return gdk_pixbuf_scale_simple(image, width, height, GDK_INTERP_BILINEAR);
} /* </pixbuf_scale> */

/*
 * scale a GdkPixmap to given width and height
 */
GdkPixbuf *
pixmap_scale(GdkPixmap *drawable, gint width, gint height)
{
  GdkPixbuf *pixbuf;
  gint orgw, orgh;

  gdk_drawable_get_size(drawable, &orgw, &orgh);
  pixbuf = gdk_pixbuf_get_from_drawable(NULL, drawable,
                                        gtk_widget_get_default_colormap(),
                                        0, 0, 0, 0, orgw, orgh);

  return pixbuf_scale(pixbuf, width, height);
} /* </pixmap_scale> */

/*
 * redraw pixbuf of a canvas drawing area
 */
void
redraw_pixbuf (GtkWidget *canvas, GdkPixbuf *image)
{
  gint xsize, ysize;

  /* Obtain width and height of the canvas area. */
  gdk_drawable_get_size (canvas->window, &xsize, &ysize);
  gdk_window_clear_area (canvas->window, 0, 0, xsize, ysize);

  if (image) {
    GdkPixbuf *pixbuf;
    gboolean   scaled;	/* we may create a new scaled pixbuf */

    gint width  = gdk_pixbuf_get_width (image);
    gint height = gdk_pixbuf_get_height (image);
    gint xorg, yorg;                            /* start (x,y) coordinates */

    float factor = fmax((float)width / xsize, (float)height / ysize) * 0.99;

    /* Adjust width and height by scale factor. */
    if (width > xsize || height > ysize) {
      width  = (float)width  / factor;
      height = (float)height / factor;
      pixbuf = pixbuf_scale (image, width, height);
      scaled = TRUE;
    }
    else {
      pixbuf = image;
      scaled = FALSE;
    }

    /* Adjust image (x,y) start coordinates. */
    xorg = abs(xsize - width) / 2;
    yorg = abs(ysize - height) / 2;

    /* Draw scaled pixbuf onto the canvas area. */
    gdk_draw_pixbuf (canvas->window,
                     canvas->style->fg_gc[GTK_WIDGET_STATE (canvas)],
                     pixbuf, 0, 0, xorg, yorg, width, height,
                     GDK_RGB_DITHER_NONE, 0, 0);
    if (scaled)
      gdk_pixbuf_unref (pixbuf);
  }
} /* </refresh> */

/*
 * set_background_pixbuf - set background pixmap of window using pixbuf
 */
void
set_background_pixbuf (GdkWindow *window, GdkPixbuf *pixbuf)
{
  GdkGC     *gc;
  GdkPixbuf *image;
  GdkPixmap *pixmap;
  gint width, height;
  gint xres, yres;


  /* Get the window and image sizes */
  gdk_window_get_size (window, &xres, &yres);
  height = gdk_pixbuf_get_height (pixbuf);
  width  = gdk_pixbuf_get_width (pixbuf);

  /* Scale GdkPixbuf *image as needed */
  if (width != xres || height != yres)
    image = pixbuf_scale (pixbuf, xres, yres);
  else
    image = pixbuf;

  /* Create pixmap at window size and render image */
  gc = gdk_gc_new (window);
  pixmap = gdk_pixmap_new (window, xres, yres, gdk_visual_get_best_depth());
  gdk_pixbuf_render_to_drawable (image, pixmap, gc, 0, 0, 0, 0, xres, yres,
                                 GDK_RGB_DITHER_NORMAL, 1, 1);
  gdk_gc_unref (gc);

  /* Set background pixmap on window */
  gdk_window_set_back_pixmap (window, pixmap, FALSE);
  gdk_pixmap_unref (pixmap);

  gdk_window_clear(window);   /* clear window to effect change */
  gdk_flush();
} /* </set_background_pixbuf> */
